#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""
This module defines :class:`~gpxity.Activity`
"""

from math import asin, sqrt, degrees
import datetime
from contextlib import contextmanager


import gpxpy
from gpxpy.gpx import GPX, GPXTrack, GPXTrackSegment



__all__ = ['Activity']


class Activity:

    """Represents an activity.

    An activity is essentially a GPX file. If a backend supports attributes not directly
    supported by the GPX format like the MapMyTracks activity type, they will
    transparently be encodeded in existing GPX fields like keywords, see :attr:`keywords`.

    The GPX part is done by https://github.com/tkrajina/gpxpy.

    If an activity is assigned to a backend, all changes will by default be written directly to the backend.
    Some backends are able to change only one attribute with little time overhead, others always have
    to rewrite the entire activity.

    You can use the context manager :meth:`batch_changes`. This holds back updating the backend until
    the context is exiting.

    Not all backends support everything, you could get the exception NotImplementedError.


    Args:
        backend (Backend): The Backend where this Activity lives in. See :attr:`backend`.
        id_in_backend (str): The identifier of this activity in the backend. See :attr:`id_in_backend`.
        gpx (GPX): Initial content.

    At least one of **backend** or **gpx** must be None. If backend is None and this
    activity is later coupled with a backend, it is silently assumed that the activity did
    not yet exist in that backend. That makes a difference when loading data from
    the backend: If data is expected in the backend but not found, an exception is raised. If
    no data is expected, we do not try to load data.

    The data will only be loaded from the backend when it is needed. Some backends
    might support loading some attributes separately, but for now, we always load
    everything as soon as anything is needed.

    Attributes:
        legal_what (tuple(str)): The legal values for :attr:`~Activity.what`. The first one is used
            as default value.

            Currently those are the values as defined by mapmytracks.
            This should eventually become more flexible.
        id_in_backend (str): Every backend has its own scheme for unique activity ids. Some
            backends may change the id if the activity data changes.
    """

    # pylint: disable = too-many-instance-attributes

    legal_what = (
        'Cycling', 'Running', 'Mountain biking', 'Indoor cycling', 'Sailing', 'Walking', 'Hiking',
        'Swimming', 'Driving', 'Off road driving', 'Motor racing', 'Motorcycling', 'Enduro',
        'Skiing', 'Cross country skiing', 'Canoeing', 'Kayaking', 'Sea kayaking', 'Stand up paddle boarding',
        'Rowing', 'Windsurfing', 'Kiteboarding', 'Orienteering', 'Mountaineering', 'Skating',
        'Skateboarding', 'Horse riding', 'Hang gliding', 'Gliding', 'Flying', 'Snowboarding',
        'Paragliding', 'Hot air ballooning', 'Nordic walking', 'Snowshoeing', 'Jet skiing', 'Powerboating',
        'Miscellaneous')

    def __init__(self, backend=None, id_in_backend: str = None, gpx=None):
        self._loading = False
        self._loaded = backend is None or id_in_backend is None
        self.__dirty = set()
        self._batch_changes = False
        self.__what = self.legal_what[0]
        self.__public = False
        self.id_in_backend = id_in_backend
        self.__gpx = gpx or GPX()
        self.__backend = None
        if backend is not None:
            if gpx is not None:
                raise Exception('Cannot accept backend and gpx')
        self.__backend = backend
        if backend and self not in backend.activities:
            backend.activities.append(self)

    @property
    def backend(self):
        """The backend this activity lives in. If it was constructed in memory, backend is None.
        If you change it from None to a backend, this activity is automatically saved in that backend.

        It is not possible to decouple an activity from its backend, use :meth:`~gpxity.activity.Activity.clone()`.

        It is not possible to move the activity to a different backend by changing this.
        Use :meth:`Backend.save() <gpxity.backend.Backend.save()>` instead.
        """
        return self.__backend

    @backend.setter
    def backend(self, value):
        if value is not self.__backend:
            if value is None:
                raise Exception('You cannot decouple an activity from its backend. Use clone().')
            elif self.__backend is not None:
                raise Exception(
                    'You cannot assign the activity to a different backend this way. '
                    'Please use Backend.save(activity).')
            else:
                self._loaded = True
                self.__backend = value
                try:
                    self.__backend.save(self)
                except BaseException:
                    self.__backend = None
                    raise

    @property
    def dirty(self) ->bool:
        """
        Is the activity in sync with the backend?

        After directly manipulating :attr:`gpx`, set dirty to True.
        See also :meth:`~gpxity.activity.Activity.gpx`.

        Setting dirty to True will directly call :meth:`~gpxity.activity.Activity._save`.

        Setting dirty to False is not allowed.

        Returns:
            bool: True if the activity is not in sync with the backend. If no backend is
            associated, False stands for an empty activity.
        """
        return bool(self.__dirty)

    @dirty.setter
    def dirty(self, value):
        if not value:
            raise Exception('You may not set dirty to False. Instead use _save().')
        if self._loading:
            return

        if isinstance(value, bool):
            self.__dirty = set(['all'])
        else:
            self.__dirty.add(value)
        self._save()

    def clone(self):
        """Creates a new activity with the same content but without backend.

        Returns:
            the new activity
        """
        result = Activity(gpx=self.__gpx.clone())
        result.what = self.what
        result.public = self.public
        return result

    def _save(self):
        """Saves all changes in the associated backend.

        If any of those conditions is met, do nothing:

        - we are currently loading from backend: Avoid recursion
        - batch_changes is active
        - we have no backend

        Otherwise asks the backend to save this activity :meth:`Backend.save() <gpxity.backend.Backend.save>`.
        """
        if self.__dirty:
            if self.backend and not self._loading and not self._batch_changes:
                self.backend.save(self, self.__dirty) # pylint: disable=no-member
                self.__dirty = set()

    @property
    def time(self) ->datetime.datetime:
        """datetime.datetime: start time of activity.
        For a simpler implementation of backends, notably MMT, we ignore
        gpx.time. Instead we return the time of the earliest track point.
        Only if there is no track point, return gpx.time.

        For the same reason time is readonly.
        """
        return self.__gpx.get_time_bounds()[0] or self.__gpx.time

    @property
    def title(self) -> str:
        """str: The title.
        """
        self._load_full()
        return self.__gpx.name

    @title.setter
    def title(self, value: str):
        if value != self.title:
            self.__gpx.name = value
            self.dirty = 'title'

    @property
    def description(self) ->str:
        """str: The description.
        """
        self._load_full()
        return self.__gpx.description

    @contextmanager
    def loading(self):
        """This context manager marks the activity as being loaded. In
        that state, automatic writes of changes into the backend are
        disabled.

        You should never need this unless you write a new backend.
        """
        prev_loading = self._loading
        self._loading = True
        try:
            yield
        finally:
            self._loading = prev_loading

    @property
    def is_loading(self):
        """True if we are currently loading. See :meth:`loading`."""
        return self._loading

    @contextmanager
    def batch_changes(self):
        """This context manager disables  the direct update in the backend
        and saves the entire activity when done.
        """
        prev_batch_changes = self._batch_changes
        self._batch_changes = True
        try:
            yield
            self._save()
        finally:
            self._batch_changes = prev_batch_changes
            self._save()

    @description.setter
    def description(self, value: str):
        if value != self.description:
            self.__gpx.description = value
            self.dirty = 'description'

    @property
    def what(self) ->str:
        """str: What is this activity doing? If we have no current value,
        return the default.

        Returns:
            The current value or the default value (see :attr:`legal_what`)
        """
        self._load_full()
        return self.__what

    @what.setter
    def what(self, value: str):
        if value != self.what:
            if value not in Activity.legal_what and value is not None:
                raise Exception('What {} is not known'.format(value))
            self.__what = value if value else self.legal_what[0]
            self.dirty = 'what'

    def _load_full(self) ->None:
        """Loads the full track from source_backend if not yet loaded."""
        if self.backend and self.id_in_backend and not self._loaded and not self._loading:
            self.backend.load_full(self) # pylint: disable=no-member
            self._loaded = True

    def add_points(self, points) ->None:
        """Adds points to last segment in the last track. If no track
        is allocated yet, do so.

        UNFINISHED

        Args:
            points (list(GPXTrackPoint): The points to be added
        """
        if self.__gpx.tracks:
            # make sure the same points are not added twice
            assert points != self.__gpx.tracks[-1].segments[-1].points[-len(points):]
        self._load_full()
        if not self.__gpx.tracks:
            self.__gpx.tracks.append(GPXTrack())
            self.__gpx.tracks[0].segments.append(GPXTrackSegment())
        self.__gpx.tracks[-1].segments[-1].points.extend(points)

    def _parse_keywords(self):
        """self.keywords is 1:1 as parsed from xml. Here we extract
        our special keywords What: and Status:"""
        new_keywords = list()
        for keyword in self.keywords:
            if keyword.startswith('What:'):
                self.what = keyword.split(':')[1]
            elif keyword.startswith('Status:'):
                self.public = keyword.split(':')[1] == 'public'
            else:
                new_keywords.append(keyword)
        self.keywords = new_keywords

    def parse(self, indata):
        """Parses GPX.

        :attr:`title`, :attr:`description` and :attr:`what` from indata have precedence over the current values.
        :attr:`public` will be or-ed

        Args:
            indata: may be a file descriptor or str
        """
        if hasattr(indata, 'read'):
            indata = indata.read()
        if not indata:
            # ignore empty file
            return
        with self.loading():
            old_gpx = self.__gpx
            old_public = self.public
            self.__gpx = gpxpy.parse(indata)
            self._parse_keywords()
            self.public = self.public or old_public
            if old_gpx.name and not self.__gpx.name:
                self.__gpx.name = old_gpx.name
            if old_gpx.description and not self.__gpx.description:
                self.__gpx.description = old_gpx.description
            self._loaded = True

    def to_xml(self) ->str:
        """Produces exactly one line per trackpoint for easier editing
        (like removal of unwanted points).
        """
        self._load_full()
        new_keywords = self.keywords
        new_keywords.append('What:{}'.format(self.what))
        if self.public:
            new_keywords.append('Status:public')
        old_keywords = self.__gpx.keywords
        try:
            self.__gpx.keywords = ', '.join(new_keywords)

            result = self.__gpx.to_xml()
            result = result.replace('</trkpt><', '</trkpt>\n<')
            result = result.replace('<link ></link>', '')   # and remove those empty <link> tags
            result = result.replace('\n</trkpt>', '</trkpt>')
            result = result.replace('\n\n', '\n')
        finally:
            self.__gpx.keywords = old_keywords
        return result

    @property
    def public(self):
        """
        bool: Is this a private activity (can only be seen by the account holder) or
            is it public?
        """
        self._load_full()
        return self.__public

    @public.setter
    def public(self, value):
        """Stores this flag as keyword 'public'."""
        if value != self.public:
            self.__public = value
            self.dirty = 'public'

    @property
    def gpx(self) ->GPX:
        """
        Direct access to the GPX object. If you use it to change its content,
        remember to set :attr:`dirty` to True afterwards.

        Returns:
            the GPX object
        """
        self._load_full()
        return self.__gpx

    def last_time(self) ->datetime.datetime:
        """
        Returns:
            the last timestamp we received so far."""
        self._load_full()
        return self.__gpx.get_time_bounds().end_time

    @property
    def keywords(self):
        """list(str): represents them as a list - in GPX they are comma separated.
            Content is whatever you want.

            Because the GPX format does not have attributes for everything used by all backends,
            we encode some of the backend arguments in keywords.

            Example for mapmytracks: keywords = 'Status:public, What:Cycling'.

            However this is transparent for you. When parsing theGPX file, those are removed
            from keywords, and the are re-added in when exporting in :meth:`to_xml`.
        """
        self._load_full()
        if self.__gpx.keywords:
            return list(x.strip() for x in self.__gpx.keywords.split(','))
        return list()

    @keywords.setter
    def keywords(self, value):
        """Replaces all keywords with a new list.

        Args:
            value (list(str)): a list of keywords
        """
        self._load_full()
        with self.batch_changes():
            self.__gpx.keywords = ''
            for keyword in value:
                # add_keyword ensures we do not get unwanted things like What:
                self.add_keyword(keyword)

    @staticmethod
    def _check_keyword(keyword):
        """Must not be What: or Status:"""
        if keyword.startswith('What:'):
            raise Exception('Do not use this directly,  use Activity.what')
        if keyword.startswith('Status:'):
            raise Exception('Do not use this directly,  use Activity.public')

    def add_keyword(self, value: str) ->None:
        """Adds to the comma separated keywords. Duplicate keywords are not allowed.

        Args:
            value: the keyword
        """
        self._check_keyword(value)
        self._load_full()
        if value in self.keywords:
            raise Exception('Keywords may not be duplicate: {}'.format(value))
        if self.__gpx.keywords:
            self.__gpx.keywords += ', {}'.format(value)
        else:
            self.__gpx.keywords = value
        self.dirty = 'keywords'

    def remove_keyword(self, value: str) ->None:
        """Removes from the keywords.

        Args:
            value: the keyword to be removed
        """
        self._check_keyword(value)
        self._load_full()
        self.__gpx.keywords = ', '.join(x for x in self.keywords if x != value)
        self.dirty = 'keywords'

    def __repr__(self):
        parts = []
        if self.backend:
            parts.append(repr(self.backend))
            parts.append('id:{}'.format(self.id_in_backend))
        if self.__gpx:
            parts.append(self.what)
            if self.__gpx.name:
                parts.append(self.__gpx.name)
            if self.__gpx.get_time_bounds()[0]:
                parts.append('{}-{}'.format(*self.__gpx.get_time_bounds()))
            parts.append('{} points'.format(self.gpx.get_track_points_no()))
            if self.angle():
                parts.append('angle={}'.format(self.angle()))
        return 'Activity({})'.format(' '.join(parts))

    def __str__(self):
        return self.__repr__()

    def key(self) ->str:
        """For speed optimized equality checks, not granted to be exact.

        Returns:
            a string with selected attributes in printable form.
        """
        self._load_full()
        return 'title:{} description:{} keywords:{} what:{}: public:{} last_time:{} angle:{} points:{}'.format(
            self.title, self.description,
            ','.join(self.keywords), self.what, self.public, self.last_time(),
            self.angle(), self.gpx.get_track_points_no())

    def angle(self) ->float:
        """For me, the earth is flat.

        Returns:
            the angle in degrees 0..360 between start and end.
            If we have no track, return 0
        """
        self._load_full()
        if not self.__gpx.tracks:
            return 0
        first_point = self.__gpx.tracks[0].segments[0].points[0]
        last_point = self.__gpx.tracks[-1].segments[-1].points[-1]
        delta_lat = first_point.latitude - last_point.latitude
        delta_long = first_point.longitude - last_point.longitude
        norm_lat = delta_lat / 90.0
        norm_long = delta_long / 180.0
        try:
            result = degrees(asin(norm_long / sqrt(norm_lat**2 + norm_long **2)))
        except ZeroDivisionError:
            return 0
        if norm_lat >= 0.0:
            return (360.0 + result) % 360.0
        return 180.0 - result

    def all_points(self):
        """
        Yields:
            GPXTrackPoint: all points in all tracks and segments
        """
        self._load_full()
        for track in self.__gpx.tracks:
            for segment in track.segments:
                for point in segment.points:
                    yield point

    def points_equal(self, other) ->bool:
        """
        Returns:
            True if both activities have identical points.

        All points of all tracks and segments are combined.
        """
        self._load_full()
        if self.gpx.get_track_points_no() != other.gpx.get_track_points_no():
            return False
        if self.angle() != other.angle():
            return False
        for _, (point1, point2) in enumerate(zip(self.all_points(), other.all_points())):
            # GPXTrackPoint has no __eq__ and no working hash()
            # those are only the most important attributes:
            if point1.longitude != point2.longitude:
                return False
            if point1.latitude != point2.latitude:
                return False
            if point1.elevation != point2.elevation:
                return False
        return True




# Copyright (c) Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

# pylint:disable=wildcard-import, missing-docstring

from .directory import *
from .server_directory import *
from .mmt import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""
This module defines :class:`~gpxity.backend.Backend`
"""

import datetime
from inspect import getmembers, isfunction
import dis

__all__ = ['Backend']


class _ActivityList(list):

    """A list of all activities in a backend,
    allowing activity id as index like in
    backend.activities['12345']

    Args:
        content (list(Activity): Initial list of activities
    """

    def __init__(self, content=None):
        super(_ActivityList, self).__init__()
        if content is not None:
            self.extend(content)

    def __contains__(self, index) ->bool:
        return self.__getitem__(index) is not None

    def __getitem__(self, index):
        """Allows accesses like alist[a_id]"""
        for _ in self:
            if _ is index:
                return _
            if _.id_in_backend == index:
                return _
        if isinstance(index, int) and 0 <= index < len(self):
            return list.__getitem__(self, index)


class Backend:
    """A place where activities live. Something like the filesystem or
    http://mapmytracks.com.

    This can be used as a context manager. At termination, all activities
    may be removed automatically, if cleanup=True. Some concrete
    implementations may also remove the backend itself.

    Not all backends support all methods. The unsupported methods
    will raise NotImplementedError. As a convenience every backend
    has a list **supported** to be used like :literal:`if 'update' in backend.supported:`
    where `update` is the name of the method.

    Args:
        url (str): the address. May be a real URL or a directory, depending on the backend implementation.
            Every implementation may define its own default for url.
        auth (tuple(str, str)): (username, password)
        cleanup (bool): If true, :meth:`destroy` will remove all activities.

    Attributes:
        supported (set(str)): The names of all supported methods. Creating the first instance of
            the backend initializes this.
    """
    supported = None

    skip_test = False

    def __init__(self, url=None, auth=None, cleanup=False):
        super(Backend, self).__init__()
        self.activities = _ActivityList()
        self._activities_fully_listed = False
        self.url = url or ''
        if self.url and not self.url.endswith('/'):
            self.url += '/'
        self.auth = auth
        self._cleanup = cleanup

    @classmethod
    def _define_support(cls):
        """If the first thing a method does is raising NotImplementedError, it is
        marked as unsupported. Those are the default values, the implementations
        will have to refine the results.
        """
        cls.supported = set()
        for name, _ in getmembers(cls, isfunction):
            if not name.startswith('_')  or (name.startswith('_write_') and name != '_write_attribute'):
                first_instruction = next(dis.get_instructions(_.__code__))
                supported = first_instruction is None or first_instruction.argval != 'NotImplementedError'
                if supported:
                    cls.supported.add(name)

    def get_time(self) ->datetime.datetime:
        """get time from the server where backend is located as a Linux timestamp"""
        raise NotImplementedError()

    def list_all(self):
        """list all activities for this user

        Returns:
            list(Activity): all activities
        """
        return list(self._yield_activities())

    def _yield_activities(self):
        """A generator for all activities. It yields the next found and appends it to activities.
        It first clears self.activities.

        Yields:
            the next activity
        """
        raise NotImplementedError()

    def list_activities(self):
        """A generator returning all activities. If all have already been listed,
        return their cached list. For rescanning first call
        :meth:`self.activities.clear()`.

        Yields:
            the next activity"""
        if not self.activities:
            self._activities_fully_listed = False
        if self._activities_fully_listed:
            for _ in self.activities:
                yield _
        else:
            for _ in self._yield_activities():
                yield _
            self._activities_fully_listed = True

    def load_full(self, activity) ->None:
        """fills the activity with all its data from source"""
        raise NotImplementedError()

    def save(self, activity, attributes=None):
        """save full activity.

        Args:
            activity (Activity): The activity we want to save in this backend.
                It may be associated with an arbitrary backend.
            attributes (set(str)): If given and the backend supports specific saving for all given attributes,
                save only those.
                Otherwise, save the entire activity.

        Returns:
            Activity: The saved activity. If the original activity lives in a different
            backend, a new activity living in this backend will be created
            and returned.
        """

        if activity.is_loading:
            raise Exception('A backend cannot save() if activity.is_loading. This is a bug in gpxity.')
        if activity.backend is not self and activity.backend is not None:
            activity = activity.clone()
        if activity.backend is None:
            activity.backend = self
            # this calls us again!
            return activity

        fully = False
        if attributes is None or attributes == set(['all']):
            fully = True
        else:
            for attribute in attributes:
                write_name = '_write_{}'.format(attribute)
                if write_name not in self.supported:
                    fully = True
                    break

        if fully:
            self._save_full(activity)
        else:
            for attribute in attributes:
                write_name = '_write_{}'.format(attribute)
                getattr(self, write_name)(activity)
        if activity not in self.activities:
            self.activities.append(activity)
        return activity

    def _save_full(self, activity) ->None:
        """the actual implementation for the concrete Backend"""
        raise NotImplementedError()

    def remove(self, activity) ->None:
        """Removes activity."""
        self._remove_activity_in_backend(activity)
        self.activities.remove(activity)
        activity.id_in_backend = None

    def _remove_activity_in_backend(self, activity) ->None:
        """backend dependent implementation"""
        raise NotImplementedError()

    def update(self, activity, points) ->None: # pylint: disable=no-self-use
        """Appends to the remove activity. points are already
        added to activity

        Todo:
            should not be exposed."""
        raise NotImplementedError()

    def remove_all(self):
        """Removes all activities we know about. If their :attr:`id_in_backend`
        has meanwhile been changed through another backend instance
        or another process, we cannot find it anymore. We do **not**
        relist all activities in the backend. If you want to make shure it
        will be empty, call :meth:`self.activities.clear` before :meth:`remove_all`."""
        for activity in list(self.list_activities()):
            self.remove(activity)

    def copy_all_from(self, from_backend):
        """Copies all activities into this backend.

        Args:
            from_backend (Backend): The source of the activities

        Returns:
            List of all new activities in this backend
        """
        result = list()
        for activity in from_backend.list_all():
            result.append(self.save(activity))
        return result

    def destroy(self):
        """If `cleanup` was set at init time, removes all activities. Some backends
       (example: :class:`Directory <gpxity.backends.directory.Directory.destroy>`)
       may also remove the account (or directory). See also :meth:`remove_all`."""
        if self._cleanup:
            self.remove_all()

    def has_same_activities(self, other) ->bool:
        """True if both backends have the same activities."""
        return set(x.key() for x in self.activities) == set(x.key() for x in other.activities)

    def __repr__(self):
        result = '{}({} {})'.format(
            self.__class__.__name__, self.url, self.auth[0] if self.auth else 'anonymous')
        return result

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, trback):
        self.destroy()

    def __eq__(self, other):
        """compares activities in both backends. Used for testing."""
        return self.has_same_activities(other)



# -*- coding: utf-8 -*-

# Copyright (c) Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""
implements test classes for Activity. They only use backend Directory,
so there is no network traffic involved (unless Directory is a network
file system, of course).
"""

import os
import io
import filecmp
import tempfile

from gpxpy.gpx import GPX

from .basic import BasicTest
from ... import Activity
from .. import Directory

# pylint: disable=attribute-defined-outside-init


class ActivityTests(BasicTest):

    """activity tests"""

    def test_init(self):
        """test initialisation"""
        Activity()
        with Directory(cleanup=True) as backend:
            with self.assertRaises(Exception):
                Activity(backend, gpx=GPX())
            Activity(backend)
            self.assertEqual(len(backend.activities), 1)

        test_url = tempfile.mkdtemp(prefix=Directory.prefix)
        self.assertTrue(os.path.exists(test_url))
        os.rmdir(test_url)
        self.assertFalse(os.path.exists(test_url))
        try:
            with Directory(url=test_url, cleanup=True):
                self.assertTrue(os.path.exists(test_url))
        finally:
            os.rmdir(test_url)

    def test_dirty(self):
        """Activity.dirty"""
        with Directory(cleanup=True) as directory:
            activity = Activity(directory)
            with self.assertRaises(Exception):
                activity.dirty = False
            self.assertFalse(activity.dirty)
            activity.dirty = True
            self.assertFalse(activity.dirty)
            activity.title = 'new title'
            self.assertFalse(activity.dirty)
            with activity.batch_changes():
                activity.title = 'new 2'
                self.assertTrue(activity.dirty)
            self.assertFalse(activity.dirty)

    def test_activity_list(self):
        """test ActivityList"""
        with Directory(cleanup=True) as directory: # we have no direct access to class ActivityList
            ali = directory.activities
            self.assertEqual(len(ali), 0)
            activity1 = Activity(backend=directory)
            self.assertIn(activity1, ali)
            self.assertIsNone(activity1.id_in_backend)
            activity1.description = 'x'
            self.assertIsNotNone(activity1.id_in_backend)

    def test_clone(self):
        """is the clone identical?"""
        activity1 = self.create_test_activity()
        activity2 = activity1.clone()
        self.assertEqualActivities(activity1, activity2)
        count1 = activity1.gpx.get_track_points_no()
        del activity1.gpx.tracks[0].segments[0].points[0]
        self.assertEqual(count1, activity1.gpx.get_track_points_no() + 1)
        self.assertNotEqualActivities(activity1, activity2)
        activity2 = activity1.clone()
        activity2.gpx.tracks[-1].segments[-1].points[-1].latitude = 5
        self.assertNotEqualActivities(activity1, activity2)
        activity2 = activity1.clone()
        activity2.gpx.tracks[-1].segments[-1].points[-1].longitude = 5
        self.assertNotEqual(activity1, activity2)
        activity2 = activity1.clone()
        last_point2 = activity2.gpx.tracks[-1].segments[-1].points[-1]
        last_point2.elevation = 500000
        self.assertEqual(last_point2.elevation, 500000)
        self.assertEqual(activity2.gpx.tracks[-1].segments[-1].points[-1].elevation, 500000)
        # here assertNotEqualActivities is wrong because keys() are still identical
        self.assertFalse(activity1.points_equal(activity2))
        activity1.gpx.tracks.clear()
        activity2.gpx.tracks.clear()
        self.assertEqualActivities(activity1, activity2)

    def test_no_what(self):
        """what must return default value if not present in gpx.keywords"""
        what_default = Activity.legal_what[0]
        activity = Activity()
        self.assertEqual(activity.what, what_default)
        activity.what = None
        self.assertEqual(activity.what, what_default)
        with self.assertRaises(Exception):
            activity.what = 'illegal value'
        self.assertEqual(activity.what, what_default)
        with self.assertRaises(Exception):
            activity.add_keyword('What:illegal value')
        self.assertEqual(activity.what, what_default)

    def test_duplicate_what(self):
        """try to add two whats to Activity"""
        what_other = Activity.legal_what[5]
        activity = Activity()
        activity.what = what_other
        with self.assertRaises(Exception):
            activity.add_keyword('What:{}'.format(what_other))

    def test_remove_what(self):
        """remove what from Activity"""
        what_default = Activity.legal_what[0]
        what_other = Activity.legal_what[5]
        activity = Activity()
        activity.what = what_other
        self.assertEqual(activity.what, what_other)
        activity.what = None
        self.assertEqual(activity.what, what_default)

    def test_no_public(self):
        """public must return False if not present in gpx.keywords"""
        activity = Activity()
        self.assertFalse(activity.public)

    def test_duplicate_public(self):
        """try to set public via its property and additionally with add_keyword"""
        activity = Activity()
        activity.public = True
        self.assertTrue(activity.public)
        with self.assertRaises(Exception):
            activity.add_keyword('Status:public')

    def test_remove_public(self):
        """remove and add public from Activity using remove_keyword and add_keyword"""
        activity = Activity()
        activity.public = True
        with self.assertRaises(Exception):
            activity.remove_keyword('Status:public')
        self.assertTrue(activity.public)
        with self.assertRaises(Exception):
            activity.add_keyword('Status:public')
        self.assertTrue(activity.public)

    def test_first_time(self):
        """about activity.time"""
        activity = self.create_test_activity()
        first_time = activity.gpx.get_time_bounds()[0]
        self.assertEqual(activity.time, first_time)

    def test_last_time(self):
        """Activity.last_time()"""
        activity = self.create_test_activity()
        gpx_last_time = activity.gpx.tracks[-1].segments[-1].points[-1].time
        self.assertEqual(activity.last_time(), gpx_last_time)

    def test_xml(self):
        """roughly check if we have one line per trackpoint"""
        activity = self.create_test_activity()
        xml = activity.to_xml()
        self.assertNotIn('<link ></link>', xml)
        lines = xml.split('\n')
        self.assertTrue(len(lines) >= activity.gpx.get_track_points_no())

    def test_parse(self):
        """does Activity parse xml correctly"""
        activity = self.create_test_activity()
        activity.keywords = ['Here are some keywords']
        xml = activity.to_xml()
        activity2 = Activity()
        activity2.parse(None)
        activity2.parse('')
        activity2.parse(xml)
        self.assertEqualActivities(activity, activity2)
        self.assertEqual(activity.keywords, activity2.keywords)
        activity2 = Activity()
        activity2.parse(io.StringIO(xml))
        self.assertEqualActivities(activity, activity2)

    def test_combine(self):
        """combine values in activity with newly parsed"""
        # first, does it overwrite?
        activity = self.create_test_activity()
        xml = activity.to_xml()
        if activity.what == 'Cycling':
            other_what = 'Running'
        else:
            other_what = 'Cycling'

        activity2 = Activity()
        activity2.title = 'Title2'
        activity2.description = 'Description2'
        activity2.what = other_what
        activity2.public = True
        activity2.parse(xml)
        self.assertEqual(activity2.title, activity.title)
        self.assertEqual(activity2.description, activity.description)
        self.assertEqual(activity2.what, activity.what)
        self.assertTrue(activity2.public)
        self.assertEqual(activity2.keywords, list())

        activity.public = True
        xml = activity2.to_xml()
        self.assertIn('Status:public', xml)
        activity2 = Activity()
        activity2.what = Activity.legal_what[3]
        activity2.public = False
        activity2.parse(xml)
        self.assertTrue(activity2.public)

        # second, does it keep old values if there are no new values?
        activity = self.create_test_activity()
        activity.title = ''
        activity.description = ''
        xml = activity.to_xml()
        if activity.what == 'Cycling':
            other_what = 'Running'
        else:
            other_what = 'Cycling'

        activity2 = Activity()
        activity2.title = 'Title2'
        activity2.description = 'Description2'
        activity2.parse(xml)
        self.assertEqual(activity2.title, 'Title2')
        self.assertEqual(activity2.description, 'Description2')

    def test_save(self):
        """save locally"""
        with Directory(cleanup=True) as directory:
            os.chmod(directory.url, 0o555)
            activity = self.create_test_activity()
            with self.assertRaises(OSError):
                activity.backend = directory
            self.assertIsNone(activity.backend)
            os.chmod(directory.url, 0o755)
            activity.backend = directory
            self.assertIsNotNone(activity.backend)

        with Directory(cleanup=True) as directory:
            dir2 = self.clone_backend(directory)
            activity = self.create_test_activity()
            activity.backend = directory
            self.assertEqual(len(directory.activities), 1)
            self.assertEqual(len(directory.activities), 1)
            aclone = activity.clone()
            self.assertEqualActivities(activity, aclone)

            self.assertEqual(len(dir2.activities), 0)
            dir2.list_all()
            self.assertEqual(len(dir2.activities), 1)

            activity2 = activity.clone()
            self.assertEqualActivities(activity, activity2)
            activity2.backend = directory
            self.assertEqual(len(directory.activities), 2)
            with self.assertRaises(Exception):
                activity2.backend = dir2
            with self.assertRaises(Exception):
                activity2.backend = None
            activity3 = dir2.save(activity2)
            self.assertEqualActivities(activity, activity3)
            self.assertEqualActivities(activity2, activity3)
            self.assertIs(activity.backend, directory)
            self.assertIs(activity2.backend, directory)
            self.assertIs(activity3.backend, dir2)
            self.assertEqual(len(directory.activities), 2)
            self.assertEqual(len(dir2.activities), 2)
            directory.list_all()
            self.assertEqual(len(directory.activities), 3)
            trunk = os.path.join(directory.url, 'Random GPX # 0')
            expected_names = list(trunk + x for x in ('.gpx', '.1.gpx', '.2.gpx'))
            files = list(os.path.join(directory.url, x) for x in os.listdir(directory.url) if x.endswith('.gpx'))
            self.assertEqual(files, expected_names)
            filecmp.clear_cache()
            for idx1, idx2 in ((0, 1), (0, 2)):
                file1 = files[idx1]
                file2 = files[idx2]
                self.assertTrue(filecmp.cmp(file1, file2),
                                'Files are different: {} and {}'.format(file1, file2))

    def test_add_points(self):
        """test Activity.add_points"""
        point_count = 11
        activity = Activity()
        points = self.some_random_points(count=point_count)
        activity.add_points(points)
        self.assertEqual(activity.gpx.get_track_points_no(), point_count)
        with self.assertRaises(Exception):
            activity.add_points(points)
        activity.add_points(points[:-1])
        self.assertEqual(activity.gpx.get_track_points_no(), point_count * 2 - 1)

    def test_points_equal(self):
        """test Activity.points_equal"""
        for _ in range(100):
            point_count = 7
            points = self.some_random_points(count=point_count)
            activity1 = Activity()
            activity1.add_points(points)
            activity2 = activity1.clone()
            points2 = list(activity2.all_points()) # those are cloned points
            self.assertTrue(activity1.points_equal(activity2))
            activity2.gpx.tracks.clear()
            activity2.add_points(points2[:5])
            self.assertFalse(activity1.points_equal(activity2))
            activity2.add_points(points2[5:])
            self.assertTrue(activity1.points_equal(activity2))

            old_long = activity2.gpx.tracks[-1].segments[-1].points[-2].longitude
            activity2.gpx.tracks[-1].segments[-1].points[-2].longitude += 1
            self.assertFalse(activity1.points_equal(activity2))
            activity2.gpx.tracks[-1].segments[-1].points[-2].longitude = old_long
            self.assertTrue(activity1.points_equal(activity2))

            old_lat = activity2.gpx.tracks[-1].segments[-1].points[-2].latitude
            activity2.gpx.tracks[-1].segments[-1].points[-2].latitude += 1
            self.assertFalse(activity1.points_equal(activity2))
            activity2.gpx.tracks[-1].segments[-1].points[-2].latitude = old_lat
            self.assertTrue(activity1.points_equal(activity2))

            activity2.gpx.tracks[-1].segments[-1].points[-2].elevation += 1
            self.assertFalse(activity1.points_equal(activity2))
            activity2.gpx.tracks[-1].segments[-1].points[-2].elevation -= 1
            self.assertTrue(activity1.points_equal(activity2), '{} != {}'.format(
                list(activity1.all_points()), list(activity2.all_points())))

            old_long = activity2.gpx.tracks[-1].segments[-1].points[-1].longitude
            activity2.gpx.tracks[-1].segments[-1].points[-1].longitude += 1
            self.assertFalse(activity1.points_equal(activity2))
            a1_points = list(activity1.all_points())
            a2_points = list(activity2.all_points())
            a1_first = a1_points[0]
            a1_last = a1_points[-1]
            a2_first = a2_points[0]
            a2_last = a2_points[-1]
            self.assertNotEqual(
                activity1.angle(), activity2.angle(),
                'a1.first:{} a1.last:{} a2.first:{} a2.last:{}'.format(
                    a1_first, a1_last, a2_first, a2_last))
            activity2.gpx.tracks[-1].segments[-1].points[-1].longitude = old_long
            self.assertTrue(activity1.points_equal(activity2))

    def test_str(self):
        """test __str__"""
        activity = Activity()
        self.assertNotIn('id:', str(activity))
        with Directory(cleanup=True) as directory:
            activity = Activity(directory)
            activity.title = 'Title'
            activity.what = 'Running'
            activity.add_points(self.some_random_points(10))
            self.assertIn('id:', str(activity))
            self.assertIn('Title', str(activity))
            self.assertIn('Running', str(activity))
            self.assertIn(str(activity.time), str(activity))
            self.assertIn(str(activity.last_time()), str(activity))
            self.assertTrue(str(activity).startswith('Activity('))
            self.assertTrue(str(activity).endswith(')'))
            activity.add_points(self.some_random_points(count=5))
            self.assertIn('5 points', str(activity))
            self.assertIn('angle=', str(activity))

    def test_angle(self):
        """test Activity.angle"""
        for _ in range(1000):
            activity1 = Activity()
            activity1.add_points(self.some_random_points(2))
            angle = activity1.angle()
            self.assertLess(angle, 360.001)
            self.assertGreater(angle, -0.001)

        activity1 = Activity()
        activity1.add_points(self.some_random_points(2))
        first_point = None
        for point in activity1.all_points():
            if first_point is None:
                first_point = point
            else:
                point.latitude = first_point.latitude
                point.longitude = first_point.longitude
        self.assertEqual(activity1.angle(), 0)

    def test_key(self):
        """Activity.key()"""
        title = 'This is a niße title'
        description = title + ' NOT - it is the description'
        what = Activity.legal_what[3]
        public = True
        points = self.some_random_points(10)
        activity = Activity()
        activity.title = title
        activity.description = description
        activity.what = what
        activity.public = public
        activity.add_points(points)
        key = activity.key()
        self.assertIn('title:{}'.format(title), key)
        self.assertIn('description:{}'.format(description), key)
        self.assertIn('what:{}'.format(what), key)
        self.assertIn('public:True', key)
        self.assertIn('last_time:{}'.format(activity.last_time()), key)
        self.assertIn('angle:{}'.format(activity.angle()), key)
        self.assertIn('points:{}'.format(activity.gpx.get_track_points_no()), key)

    def test_symlinks(self):
        """Directory symlinks"""
        with Directory(cleanup=True) as directory:
            source = os.path.join(directory.url, 'deadlink')
            target = 'deadtarget'
            target_path = os.path.join(directory.url, target)
            with open(target_path, 'w') as target_file:
                target_file.write(' ')
            os.symlink('deadtarget', source)
            os.remove(target_path)
            with self.assertRaises(Exception):
                directory.list_all() # this loads symlinks



# -*- coding: utf-8 -*-

# Copyright (c) Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""
This module defines :class:`~gpxity.Authenticate`
"""

from pkgutil import get_data
from configparser import ConfigParser

from gpxity import Backend

__all__ = ['Authenticate']

class Authenticate:

    """
    Get username and password from auth.cfg. If nothing is
    useable, sets them to None.

    Args:
        cls (Backend): The class of the backend
        sub_name (str): Be more specific. This can be used to define different data for some tests.

    Attributes:
        auth (tuple(str,str)): (username, password). Both are either str or None.

    auth.cfg has sections
      * [default]             most general fallback
      * [ClassName]           the class name of a backend like MMT
      * [ClassName.sub_name]  can be used for a specific test

    Those sections are tried from most specific to default until
    both username and password are known. It is legal if a more
    specific section only defines username or password.

    """

    # pylint: disable=too-few-public-methods

    def __init__(self, cls: Backend, sub_name: str = None):

        self.cls = cls
        self.sub_name = sub_name
        self.auth = (None, None)

        try:
            with open('auth.cfg') as auth_file:
                self._parse_config(auth_file.read())
        except BaseException:
            try:
                self._parse_config(get_data(__package__, 'backends/test/auth.cfg').decode())
            except BaseException:
                pass

    def _parse_config(self, data):
        """try to use data"""

        username = password = None

        config = ConfigParser()
        config.read_string(data)

        # try most specific section first, default last:
        try_sections = list([self.cls.__name__, 'default'])
        if self.sub_name:
            try_sections.insert(0, (try_sections[0] + '.' + self.sub_name))

        for check_section in try_sections:
            if check_section in config.sections():
                section = config[check_section]
                if username is None and 'Username' in section:
                    username = section['Username']
                if password is None and 'Password' in section:
                    password = section['Password']
                if username and password:
                    break

        self.auth = (username, password)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""
This implements :class:`gpxity.backends.ServerDirectory`
"""


from .directory import Directory

__all__ = ['ServerDirectory']

class ServerDirectory(Directory):
    """Like :class:`Directory` but the activity ids are different: Just a number.
    A new id is generated by adding 1 to the highest existing id.

    The symbolic links per YYYY/MM use the title of the activity as link name.

    """

    # pylint: disable=abstract-method

    skip_test = False

    def _set_new_id(self, activity):
        """gives the activity a unique id"""
        try:
            activity.id_in_backend = str(max(int(x) for x in self._list_gpx()) + 1)
        except ValueError:
            activity.id_in_backend = '1'

ServerDirectory._define_support() # pylint: disable=protected-access



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""
Defines :class:`gpxity.backends.directory.Directory`
"""


import os
import datetime
import tempfile
from collections import defaultdict

from .. import Backend, Activity

__all__ = ['Directory']

class Directory(Backend):
    """Uses a directory for storage. The filename minus the .gpx ending is used as the storage id.
    If the activity has a title, use the title as storage id, making it unique by attaching a number if needed.
    An activity without title gets a random name.

    The main directory (given by :attr:`~gpxity.backends.directory.Directory.url`) will have
    subdirectories YYYY/MM (year/month) with only the activities for one month.
    Those are symbolic links to the main file and have the same file name.

    This backend uses activity.title for the id. If an activity has no title, it uses a random
    sequence of characters. Changing the title also changes the id.

    Args:
        url (str): a directory. If not given, use a unique temporary directory named
            gpxpy.X where X are some random characters.
            If the directory does not exist, it is created.
        auth (tuple(str, str)): Unused.
        cleanup (bool): If True, :meth:`destroy` will remove all activities. If url was
            not given, it will also remove the directory.

    Attributes:
        prefix (str): The prefix for temporary directories.
    """

   # skip_test = True
    # pylint: disable=abstract-method

    prefix = 'gpxity.'

    def __init__(self, url=None, auth=None, cleanup=False):
        self.url_given = bool(url)
        if not self.url_given:
            url = tempfile.mkdtemp(prefix=self.prefix)
        super(Directory, self).__init__(os.path.abspath(os.path.expanduser(url)), auth=auth, cleanup=cleanup)
        if not os.path.exists(self.url):
            os.makedirs(self.url)
        self._symlinks = None
        self._load_symlinks()

    def _load_symlinks(self):
        """scan the subdirectories with the symlinks. If the content of an
        actiivty changes, the symlinks might have to be adapted. But
        we do not know the name of the existing symlink anymore. So
        just scan them all and assign them to id_in_backend."""
        self._symlinks = defaultdict(list)
        for dirpath, _, filenames in os.walk(self.url):
            for filename in filenames:
                full_name = os.path.join(dirpath, filename)
                if os.path.islink(full_name):
                    if os.path.exists(full_name):
                        target = os.readlink(full_name)
                        gpx_target = os.path.basename(target)
                        if gpx_target.endswith('.gpx'):
                            # it really should ...
                            gpx_target = gpx_target[:-4]
                        self._symlinks[gpx_target].append(full_name)
                    else:
                        os.remove(full_name)
                        raise Exception('{}: removed dead symbolic link {}'.format(
                            self, full_name))

    def _set_new_id(self, activity):
        """a not yet existant file name"""
        if activity.title:
            value = activity.title
        else:
            value = os.path.basename(tempfile.NamedTemporaryFile(dir=self.url, prefix='').name)
        path = self._make_path_unique(os.path.join(self.url, value + '.gpx'))
        activity.id_in_backend = os.path.basename(path)[:-4]

    @staticmethod
    def _make_path_unique(value):
        """if the file name already exists, append a serial number"""
        ctr = 0
        unique_value = value
        while os.path.exists(unique_value):
            ctr += 1
            if value.endswith('.gpx'):
                unique_value = '{}.{}.gpx'.format(value[:-4], ctr)
            else:
                unique_value = '{}.{}'.format(value, ctr)
        return unique_value

    def destroy(self):
        """If `cleanup` was set at init time, removes all activities.
        If :attr:`~gpxity.backends.directory.Directory.url` was set at init time,
        also removes the directory."""
        super(Directory, self).destroy()
        if self._cleanup:
            self.remove_all()
            if not self.url_given:
                os.rmdir(self.url)

    def _gpx_path(self, activity):
        """The full path name for the local copy of an activity"""
        if not activity.id_in_backend:
            self._set_new_id(activity)
        base_name = '{}.gpx'.format(activity.id_in_backend)
        return os.path.join(self.url, base_name)

    def _list_gpx(self):
        """returns a generator of all gpx files, with .gpx removed"""
        gpx_names = (x for x in os.listdir(self.url) if x.endswith('.gpx'))
        return (x.replace('.gpx', '') for x in gpx_names)

    def _yield_activities(self):
        self.activities.clear()
        self._load_symlinks()
        for _ in self._list_gpx():
            yield Activity(self, _)

    def get_time(self) ->datetime.datetime:
        """get server time as a Linux timestamp"""
        return datetime.datetime.now()

    def load_full(self, activity):
        """fills the activity with all its data from source."""
        with activity.loading():
            with open(self._gpx_path(activity)) as in_file:
                activity.parse(in_file)

    def _remove_activity_in_backend(self, activity):
        """Removes its symlinks, empty symlink parent directories  and the file, in this order."""
        for symlink in self._symlinks[activity.id_in_backend]:
            os.remove(symlink)
            symlink_dir = os.path.split(symlink)[0]
            try:
                os.removedirs(symlink_dir)
            except OSError:
                pass
        self._symlinks[activity.id_in_backend] = list()
        gpx_file = self._gpx_path(activity)
        if os.path.exists(gpx_file):
            os.remove(gpx_file)

    def _symlink_path(self, activity):
        """The path for the speaking symbolic link: YYYY/MM/title.gpx.
        Missing directories YYYY/MM are created.
        activity.time must be set."""
        time = activity.time
        by_month_dir = os.path.join(self.url, '{}'.format(time.year), '{:02}'.format(time.month))
        if not os.path.exists(by_month_dir):
            os.makedirs(by_month_dir)
        name = activity.title or activity.id_in_backend
        return self._make_path_unique(os.path.join(by_month_dir, name))

    def _save_full(self, activity):
        """save full gpx track. Since the file name uses title and title may have changed,
        compute new file name and remove the old files. We also adapt activity.id_in_backend."""
        self._remove_activity_in_backend(activity)
        if activity.title:
            # enforce new id_in_backend using title
            activity.id_in_backend = None
        _gpx_path = self._gpx_path(activity)
        try:
            with open(_gpx_path, 'w') as out_file:
                out_file.write(activity.to_xml())
            time = activity.time
            if time:
                os.utime(_gpx_path, (time.timestamp(), time.timestamp()))
                link_name = self._symlink_path(activity)
                link_target = os.path.join('..', '..', '{}.gpx'.format(activity.id_in_backend))
                os.symlink(link_target, link_name)
                self._symlinks[activity.id_in_backend].append(link_name)
        except BaseException:
            raise

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.url)

Directory._define_support() # pylint: disable=protected-access



# -*- coding: utf-8 -*-

# Copyright (c) Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""
implements :class:`gpxpy.backends.test.test_backends.TestBackends` for all backends
"""

import time
import requests

from .basic import BasicTest
from .. import Directory, MMT, ServerDirectory
from ... import Activity

# pylint: disable=attribute-defined-outside-init


class TestBackends(BasicTest):
    """Are the :literal:`supported_` attributes set correctly?"""

    def test_supported(self):
        """Check values in supported for all backends"""
        expect_unsupported = dict()
        expect_unsupported[Directory] = set(['update'])
        expect_unsupported[ServerDirectory] = set(['update'])
        expect_unsupported[MMT] = set()
        for cls in self._find_backend_classes():
            with self.subTest(' {}'.format(cls.__name__)):
                self.assertTrue(cls.supported & expect_unsupported[cls] == set())

    def test_save_empty(self):
        """Save empty activity"""
        for cls in self._find_backend_classes():
            with self.subTest(' {}'.format(cls.__name__)):
                backend = self.setup_backend(cls, cleanup=True)
                try:
                    activity = Activity()
                    if cls is MMT:
                        with self.assertRaises(Exception):
                            backend.save(activity)
                    else:
                        self.assertIsNotNone(backend.save(activity))
                finally:
                    backend.destroy()

    def test_backend(self):
        """Manipulate backend"""
        activity = self.create_test_activity()
        with Directory(cleanup=True) as directory1:
            with Directory(cleanup=True) as directory2:
                saved = directory1.save(activity)
                self.assertEqual(saved.backend, directory1)
                activity.backend = directory1
                with self.assertRaises(Exception):
                    activity.backend = directory2
                with self.assertRaises(Exception):
                    activity.backend = None

    def test_open_wrong_auth(self):
        """Open backends with wrong password"""
        for cls in self._find_backend_classes():
            with self.subTest(' {}'.format(cls.__name__)):
                if issubclass(cls, Directory):
                    try:
                        backend = self.setup_backend(cls, sub_name='wrong', cleanup=True)
                    finally:
                        backend.destroy()
                else:
                    with self.assertRaises(requests.exceptions.HTTPError):
                        self.setup_backend(cls, sub_name='wrong')

    def test_create_backend(self):
        """Test creation of a backend"""
        for cls in self._find_backend_classes():
            with self.subTest(' {}'.format(cls.__name__)):
                backend = self.setup_backend(cls, count=3, clear_first=True, cleanup=True)
                try:
                    self.assertEqual(len(backend.list_all()), 3)
                    first_time = backend.get_time()
                    time.sleep(2)
                    second_time = backend.get_time()
                    total_seconds = (second_time - first_time).total_seconds()
                    self.assertTrue(1 < total_seconds < 4, 'Time difference should be {}, is {}-{}={}'.format(
                        2, second_time, first_time, second_time - first_time))
                finally:
                    backend.destroy()

    def test_write_remote_attributes(self):
        """If we change title, description, public, what in activity, is the backend updated?"""
        for cls in self._find_backend_classes():
            with self.subTest(' {}'.format(cls.__name__)):
                backend = self.setup_backend(cls, count=1, clear_first=True, cleanup=True)
                try:
                    activity = backend.list_all()[0]
                    first_public = activity.public
                    first_title = activity.title
                    first_description = activity.description
                    first_what = activity.what
                    activity.public = not activity.public
                    activity.title = 'A new title'
                    self.assertEqual(activity.title, 'A new title')
                    activity.description = 'A new description'
                    if activity.what == 'Cycling':
                        activity.what = 'Running'
                    else:
                        activity.what = 'Cycling'
                    # make sure there is no cache in the way
                    backend2 = self.clone_backend(backend)
                    activity2 = backend2.list_all()[0]
                    self.assertNotEqual(first_public, activity2.public)
                    self.assertNotEqual(first_title, activity2.title)
                    self.assertNotEqual(first_description, activity2.description)
                    self.assertNotEqual(first_what, activity2.what)
                finally:
                    backend.destroy()

    def test_all_what(self):
        """can we up- and download all values for :attr:`Activity.what`?"""
        what_count = len(Activity.legal_what)
        backends = list(
            self.setup_backend(x, count=what_count, clear_first=True)
            for x in self._find_backend_classes())
        copies = list(self.clone_backend(x) for x in backends)
        first_backend = copies[0]
        for other in copies[1:]:
            self.assertSameActivities(first_backend, other)
        for backend in copies:
            backend.destroy()
        for backend in backends:
            backend.destroy()

    def test_keywords(self):
        """save and load keywords"""
        for cls in self._find_backend_classes():
            with self.subTest(' {}'.format(cls.__name__)):
                backend = self.setup_backend(cls, count=1, clear_first=True, cleanup=True)
                try:
                    activity = Activity(backend)
                    activity.add_points(self.some_random_points(5))
                    activity.keywords = (['a', 'b', 'c'])
                    with self.assertRaises(Exception):
                        activity.add_keyword('b')
                    activity.remove_keyword('b')
                    self.assertEqual(activity.keywords, (['a', 'c']))
                    with self.assertRaises(Exception):
                        activity.add_keyword('What:whatever')
                    activity.add_keyword('e')
                    self.assertEqual(activity.keywords, (['a', 'c', 'e']))
                finally:
                    backend.destroy()

    def test_unicode(self):
        """Can we up- and download unicode characters in all text attributes?"""
        tstdescr = 'DESCRIPTION with utf-8 char ß (unicode szlig) and something japanese:の諸問題'
        for cls in self._find_backend_classes():
            with self.subTest(' {}'.format(cls.__name__)):
                backend = self.setup_backend(cls, count=1, clear_first=True)
                backend2 = self.clone_backend(backend)
                try:
                    activity = backend.list_all()[0]
                    activity.title = 'Title with utf-8 char ß (unicode szlig)'
                    backend2.list_all()
                    activity2 = backend2.activities[0]
                    self.assertEqualActivities(activity, activity2)
                    activity.description = tstdescr
                    self.assertEqual(activity.description, tstdescr)
                    backend2.activities.clear()
                    backend2.list_all()
                    activity2 = backend2.activities[0]
                    self.assertEqual(activity2.description, tstdescr)
                    self.assertEqualActivities(activity, activity2)
                finally:
                    backend.destroy()
                    backend2.destroy()

    def test_change_points(self):
        """Can we change the points of a track?

        For MMT this means re-uploading and removing the previous instance, so this
        is not always as trivial as it should be."""

    def test_download_many(self):
        """download many activities"""
        many = 100
        backend = self.setup_backend(MMT, count=many, cleanup=False, clear_first=False, sub_name='many')
        self.assertEqual(len(backend.list_all()), many)

    def test_duplicate_title(self):
        """two activities having the same title"""
        for cls in self._find_backend_classes():
            with self.subTest(' {}'.format(cls.__name__)):
                backend = self.setup_backend(cls, count=2, clear_first=True)
                activity1, activity2 = backend.list_all()
                try:
                    activity1.title = 'TITLE'
                    activity2.title = 'TITLE'
                finally:
                    backend.destroy()



# -*- coding: utf-8 -*-


# Copyright (c) Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""
Tests for gpxity.backends
"""

import unittest
import importlib
import os
import io
import datetime
import random
from inspect import getmembers, isclass, getmro
from pkgutil import get_data
import tempfile

import gpxpy
from gpxpy.gpx import GPXTrackPoint

from ...activity import Activity
from ...auth import Authenticate
from ...backend import Backend

# pylint: disable=attribute-defined-outside-init

__all__ = ['BasicTest']

from .. import Directory

class BasicTest(unittest.TestCase):
    """define some helpers

    Attributes:
        all_backend_classes: a list of all backend implementations
        auth (tuple(str, str)): username/password
    """

    all_backend_classes = None

 #   def __init__(self):
    #    super(BasicTest, self).__init__()

    def setUp(self):
        """defines test specific Directory.prefix"""
        Directory.prefix = 'gpxity.' + '.'.join(self.id().split('.')[-2:]) + '/'
        path = os.path.join(tempfile.gettempdir(), Directory.prefix)
        if not os.path.exists(path):
            os.mkdir(path)

    def tearDown(self):
        """Check if there are still /tmp/gpxitytest.* directories"""
        must_be_empty = os.path.join(tempfile.gettempdir(), Directory.prefix)
        os.rmdir(must_be_empty)

    def setup_auth(self, cls_, sub_name=None):
        """get auth data. Save it in self.auth

        Args:
            cls_ (Backend): The backend class
            sub_name: for more specific username/passwords, see :class:`gpxity.auth.Authenticate`
        """
        self.auth = Authenticate(cls_, sub_name).auth

    @staticmethod
    def create_test_activity(count=1, idx=0, what=None):
        """creates an activity. It starts off with **test.gpx** and appends a
        last track point, it also changes the time stamp of the last point.
        This is done using **count** and **idx**: The last point is set such that
        looking at the tracks, they all go in a different direction clockwise, with an angle
        in degrees of :literal:`360 * idx / count`.

        Args:
            count (int): See above. Using 1 as default if not given.
            idx (int): See above. Using 0 as default if not given.
            what (str): The wanted value for the activity.
                Default: if count == len(:attr:`Activity.legal_what <gpxity.activity.Activity.legal_what>`),
                the default value will be legal_what[idx].
                Otherwise a random value will be applied.

        Returns:
            (Activity): A new activity not bound to a backend
        """
        if BasicTest.all_backend_classes is None:
            BasicTest.all_backend_classes = BasicTest._find_backend_classes()
        gpx_test_file = os.path.join(os.path.dirname(__file__), 'test.gpx')
        if not os.path.exists(gpx_test_file):
            raise Exception('MMTTests needs a GPX file named test.gpx for testing in {}'.format(
                os.getcwd()))

        data = io.StringIO(get_data(__package__, 'test.gpx').decode('utf-8'))
        gpx = gpxpy.parse(data)
        movement = gpxpy.geo.LocationDelta(distance=100000, angle=360 * idx / count)
        last_points = gpx.tracks[-1].segments[-1].points
        new_point = GPXTrackPoint(
            latitude=last_points[-1].latitude, longitude=last_points[-1].longitude + 0.001,
            time=last_points[-1].time + datetime.timedelta(hours=10, seconds=idx))
        new_point.move(movement)
        gpx.tracks[-1].segments[-1].points.append(new_point)
        result = Activity(gpx=gpx)
        result.title = 'Random GPX # {}'.format(idx)
        result.description = 'Description to {}'.format(gpx.name)
        if what:
            result.what = what
        elif count == len(Activity.legal_what):
            result.what = Activity.legal_what[idx]
        else:
            result.what = random.choice(Activity.legal_what)
        return result

    @staticmethod
    def _random_datetime():
        """random datetime between now() - 10 days and now()"""
        end = datetime.datetime.now()
        start = end - datetime.timedelta(days=10)
        delta = end - start
        int_delta = (delta.days * 24 * 60 * 60) + delta.seconds
        random_second = random.randrange(int_delta)
        return start + datetime.timedelta(seconds=random_second)

    @classmethod
    def some_random_points(cls, count=10):
        """

        Returns:
            A list with count points
        """
        result = list()
        for _ in range(count):
            point = GPXTrackPoint(
                latitude=random.uniform(0.0, 90.0),
                longitude=random.uniform(0.0, 180.0), elevation=_,
                time=cls._random_datetime() + datetime.timedelta(seconds=10))
            result.append(point)
        return result

    def assertSameActivities(self, backend1, backend2): # pylint: disable=invalid-name
        """both backends must hold identical activities"""
        self.maxDiff = None # pylint: disable=invalid-name
        self.assertEqual(backend1, backend2, 'backend1:{} backend2:{}'.format(
            list(x.key() for x in backend1.activities),
            list(x.key() for x in backend2.activities)))

    def assertEqualActivities(self, activity1, activity2): # pylint: disable=invalid-name
        """both activities must be identical. We test more than necessary for better test coverage."""
        self.maxDiff = None
        self.assertEqual(activity1.key(), activity2.key())
        self.assertTrue(activity1.points_equal(activity2))
        self.assertEqual(activity1.gpx.to_xml(), activity2.gpx.to_xml())

    def assertNotEqualActivities(self, activity1, activity2): # pylint: disable=invalid-name
        """both activities must be identical. We test more than necessary for better test coverage."""
        self.assertNotEqual(activity1.key(), activity2.key())
        self.assertFalse(activity1.points_equal(activity2))
        self.assertNotEqual(activity1.gpx.to_xml(), activity2.gpx.to_xml())

    def setup_backend(self, cls_, url=None, count=0, cleanup=True, clear_first=True, sub_name=None):
        """sets up an instance of a backend with count activities.

        If count == len(:attr:`Activity.legal_what <gpxity.activity.Activity.legal_what>`),
        the list of activities will always be identical. For an example
        see :meth:`TestBackends.test_all_what <gpxity.backends.test.test_backends.TestBackends.test_all_what>`.

        Args:
            cls_ (Backend): the class of the backend to be created
            url (str): for the backend
            count (int): how many random activities should be inserted?
            cleanup (bool): If True, remove all activities when done. Passed to the backend.
            clear_first (bool): if True, first remove all existing activities
            sub_name (str): use this to get specific username/passwords from Authenticate

        Returns:
            the prepared Backend
        """

        self.setup_auth(cls_, sub_name)
        result = cls_(url, auth=self.auth, cleanup=cleanup)
        if clear_first:
            result.remove_all()
        else:
            result.list_all()
        while count > len(result.activities):
            activity = self.create_test_activity(count, len(result.activities))
            result.save(activity)
        return result

    @staticmethod
    def clone_backend(backend):
        """returns a clone of backend with nothing listed or loaded
        """
        return backend.__class__(backend.url, backend.auth)

    @staticmethod
    def _find_backend_classes(with_skip: bool = False):
        """Finds all backend classes. Those will be tested.

        Args:
            with_skip: if True, also finds those with skip_test=False
        """
        backends_directory = __file__
        while not backends_directory.endswith('backends'):
            backends_directory = os.path.dirname(backends_directory)
        if not os.path.exists(backends_directory):
            raise Exception('we are not where we should be')
        result = list()
        mod_names = os.listdir(backends_directory)
        for mod in mod_names:
            if not mod.endswith('.py'):
                continue
            mod = mod.replace('.py', '')
            if mod == '__init__':
                continue
            try:
                imported = importlib.__import__(mod, globals(), locals(), level=2)
                for name, cls in getmembers(imported, isclass):
                    if name in imported.__all__ and Backend in getmro(cls)[1:]:
                        # isinstance and is do not work here
                        if with_skip or not cls.skip_test:
                            result.append(cls)
            except ImportError:
                pass
        #sort because we want things reproducibly
        return sorted(set(result), key=lambda x: x.__class__.__name__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) Wolfgang Rohdewald <wolfgang@rohdewald.de>
# See LICENSE for details.

"""
This implements :class:`gpxity.backends.MMT` for http://www.mapmytracks.com

There are some problems with the server running at mapmytracks.com:
    * it is not possible to change an existing activity - if the track changes, the
      activity must be re-uploaded and gets a new activity id This invalididates
      references held by other backend instances (maybe on other clients).
      But I could imagine that most similar services have this problem too.
    * does not support GPX very well beyond track data. One problem is that
      it does not support gpx.time, it ignores it in uploads and uses the time
      of the earliest trackpoint. To be consistent, Gpxity follows that for now
      and does not respect gpx.time either.
    * there is an official description of an API at https://github.com/MapMyTracks
      but this does not implement everything needed. For the missing parts we
      simulate what a web browser would do, see :meth:`MMT.load_full` and
      :meth:`MMT._write_attribute`. Of course that could fail if MMT changes its site.
      Which is true for the api itself, it can and does get incompatible changes at
      any time without notice to users or deprecation periods.
    * downloading activities with that abi is very slow and hangs forever for big activities
      (at least this was so in Feb 2017, maybe have to test again occasionally).
    * not all parts of MMT data are supported like tags (would be nice to have.
      Download is implemented but no upload) or images (not interesting for me,
      at least not now).

"""

from xml.etree import ElementTree
from html.parser import HTMLParser
import datetime

import requests

from .. import Backend, Activity


__all__ = ['MMT']

def _convert_time(raw_time) ->datetime.datetime:
    """MMT uses Linux timestamps. Converts that into datetime

    Args:
        raw_time (int): The linux timestamp from the MMT server
    """
    return datetime.datetime.utcfromtimestamp(float(raw_time))


class MMTSession:
    """Helps execute commands while logged in"""
    # pylint: disable=too-few-public-methods

    def __init__(self, backend):
        self.session = requests.Session()
        # I have no idea what ACT=9 does but it seems to be needed
        payload = {'username': backend.auth[0], 'password': backend.auth[1], 'ACT':'9'}
        base_url = backend.url.replace('http:', 'https:').replace('/api/', '')
        login_url = '{}/login'.format(base_url)
        response = self.session.post(login_url, data=payload)
        if not 'You are now logged in.' in response.text:
            raise requests.exceptions.HTTPError()

    def __enter__(self):
        return self.session

    def __exit__(self, exc_type, exc_value, trback):
        self.session.close()


class ParseMMTActivity(HTMLParser): # pylint: disable=abstract-method

    """get some attributes available only on the web page. Of course,
    this is highly unreliable. Just use what we can get."""

    result = dict()

    def __init__(self):
        super(ParseMMTActivity, self).__init__()
        self.seeing_what = False
        self.seeing_title = False
        self.seeing_description = False
        self.seeing_status = False
        self.result['title'] = None
        self.result['description'] = None
        self.result['legal_whats'] = list()
        self.result['what'] = None
        self.result['what_from_title'] = None
        self.result['what_3'] = None
        self.result['public'] = None

    def handle_starttag(self, tag, attrs):
        """starttag from the parser"""
        self.seeing_title = False
        self.seeing_description = False
        self.seeing_what = False
        self.seeing_status = False
        if tag == 'input' and len(attrs) == 4:
            if attrs[0] == ('id', 'activity_type'):
                if attrs[1] == ('type', 'hidden'):
                    if attrs[2] == ('name', 'activity_type'):
                        if attrs[3][0] == 'value':
                            value = attrs[3][1].strip()
                            self.result['what_3'] = value
            if attrs[0] == ('type', 'radio') and attrs[1] == ('class', 'activity_type'):
                value = attrs[3][1].strip()
                self.result['legal_whats'].append(value)
        elif tag == 'div' and attrs and attrs[0] == ('class', 'panel') and attrs[1][0] == 'data-activity':
            # sometime this says Miscellaneous instead of the correct value like Swimming
            self.result['what'] = attrs[1][1]
        elif tag == 'span' and attrs and attrs[0] == ('class', 'privacy-status'):
            self.seeing_status = True
        elif tag == 'title':
            self.seeing_what = True
        elif tag == 'h2' and attrs and attrs[0] == ('id', 'track-title'):
            self.seeing_title = True
        elif tag == 'p' and attrs and attrs[0] == ('id', 'track-desc'):
            self.seeing_description = True

    def handle_data(self, data):
        """data from the parser"""
        if not data.strip():
            return
        if self.seeing_title:
            self.result['title'] = data.strip()
        if self.seeing_description:
            self.result['description'] = data.strip()
        if self.seeing_what:
            try:
                _ = data.split('|')[1].split('@')[0].strip()
                self.result['what_from_title'] = ' '.join(_.split(' ')[:-2])
            except BaseException:
                print('cannot parse', data)
                self.result['what_from_title'] = ''
        if self.seeing_status:
            self.result['public'] = data.strip() != 'Only you can see this activity'


class MMTRawActivity:

    """raw data from mapmytracks.get_activities"""

    # pylint: disable=too-few-public-methods
    def __init__(self, xml):
        self.activity_id = xml.find('id').text
        self.title = xml.find('title').text
        self.time = _convert_time(xml.find('date').text)
        self.what = xml.find('activity_type').text


class MMT(Backend):
    """The implementation for MapMyTracks.
    The activity ident is the number given by MapMyTracks.

    Args:
        url (str): The Url of the server. Default is http://mapmytracks.com/api
        auth (tuple(str, str)): Username and password
        cleanup (bool): If True, :meth:`~gpxity.backend.Backend.destroy` will remove all activities in the
            user account.
    """

    # pylint: disable=abstract-method

   #  skip_test = True

    def __init__(self, url=None, auth=None, cleanup=True):
        if url is None:
            url = 'http://www.mapmytracks.com/api'
        super(MMT, self).__init__(url, auth, cleanup)
        self.remote_known_whats = None

    def __post(self, request, session=None, **kwargs):
        """helper for the real function"""
        data = kwargs.copy()
        data['request'] = request
        try:
            response = (session or requests).post(self.url, data=data, auth=self.auth, timeout=(5, 300))
        except requests.exceptions.ReadTimeout:
            print('timeout for', data)
            raise
        if response.status_code != requests.codes.ok: # pylint: disable=no-member
            self.__handle_post_error(request, response)
            return
        try:
            result = ElementTree.fromstring(response.text)
        except ElementTree.ParseError:
            print('POST {} has parse error in {}: {}'.format(data, request, response.text))
            raise
        assert result.text != 'error' # should have raise_for_status
        result_type = result.find('type')
        if result_type is not None and result_type.text == 'error':
            reason = result.find('reason').text if result.find('reason') else 'no reason given'
            raise requests.exceptions.HTTPError('{}: {}'.format(data, reason))
        return result

    def __handle_post_error(self, request, result):
        """we got status_code != ok"""
        try:
            result.raise_for_status()
        except BaseException as exc:
            raise type(exc)('{}: {} {} {}'.format(exc, self.url, request, result.text))

    def _write_attribute(self, activity, attribute):
        """change an attribute directly on mapmytracks. Note that we specify iso-8859-1 but
        use utf-8. If we correctly specify utf-8 in the xml encoding, mapmytracks.com
        aborts our connection."""
        if activity.is_loading:
            return
        with MMTSession(self) as session:
            url = self._base_url() + '/assets/php/interface.php'
            data = '<?xml version="1.0" encoding="ISO-8859-1"?>' \
                '<message><nature>update_{attr}</nature><eid>{eid}</eid>' \
                '<usr>{usrid}</usr><uid>{uid}</uid>' \
                '<{attr}>{value}</{attr}></message>'.format(
                    attr=attribute,
                    eid=activity.id_in_backend,
                    usrid=self.auth[0],
                    value=getattr(activity, attribute),
                    uid=session.cookies['exp_uniqueid']).encode('ascii', 'xmlcharrefreplace')
            response = session.post(url, data=data)
            if 'success' not in response.text:
                raise requests.exceptions.HTTPError('{}: {}'.format(data, response.text))

    def _write_title(self, activity):
        """changes title on remote server"""
        self._write_attribute(activity, 'title')

    def _write_description(self, activity):
        """changes description on remote server"""
        self._write_attribute(activity, 'description')

    def _write_public(self, activity):
        """changes public/private on remote server"""
        with MMTSession(self) as session:
            url = self._base_url() + '/assets/php/interface.php'
            data = '<?xml version="1.0"?>' \
                '<message><nature>toggle_status</nature><eid>{}</eid>' \
                '<usr>{}</usr><uid>{}</uid>' \
                '</message>'.format(
                    activity.id_in_backend, self.auth[0],
                    session.cookies['exp_uniqueid']).encode('ascii')
            response = session.post(url, data=data)
            if 'success' not in response.text:
                raise requests.exceptions.HTTPError()
            wanted_public = activity.public
            # should reall be only in unittest:
            self._load_page_in_session(activity, session)
            assert activity.public == wanted_public

    def _write_what(self, activity):
        """change what directly on mapmytracks. Note that we specify iso-8859-1 but
        use utf-8. If we correctly specify utf-8 in the xml encoding, mapmytracks.com
        aborts our connection."""
        with MMTSession(self) as session:
            url = self._base_url() + '/handler/change_activity'
            data = {'eid': activity.id_in_backend, 'activity': activity.what}
            response = session.post(url, data=data)
            if 'ok' not in response.text:
                raise requests.exceptions.HTTPError('{}: {}'.format(data, response.text))

    def get_time(self) ->datetime.datetime:
        """get MMT server time"""
        return _convert_time(self.__post('get_time').find('server_time').text)

    def _yield_activities(self):
        """get all activities for this user. If we do not use the generator
        created by yield_activity, unittest fails. Why?"""

        self.activities.clear()

        with MMTSession(self) as session:
            while True:
                old_len = len(self.activities)
                response = self.__post(
                    'get_activities', author=self.auth[0],
                    offset=old_len, session=session)
                chunk = response.find('activities')
                if not chunk:
                    return
                for _ in chunk:
                    raw_data = MMTRawActivity(_)
                    activity = Activity(self, raw_data.activity_id)
                    with activity.loading():
                        activity.title = raw_data.title
                        activity.what = raw_data.what
                    yield activity
                assert len(self.activities) > old_len

    def _base_url(self):
        """the url without subdirectories"""
        return self.url.replace('/api/', '')

    def _load_page_in_session(self, activity, session):
        """The MMT api does not deliver all attributes we want.
        This gets some more by scanning the web page."""
        with activity.loading():
            response = session.get('{}/explore/activity/{}'.format(
                self._base_url(), activity.id_in_backend))
            page_parser = ParseMMTActivity()
            page_parser.feed(response.text)
            # if the title has not been set, get_activities says something like "Activity 2016-09-04 ..."
            # while the home page says "Cycling activity". We prefer the value from the home page
            # and silently ignore this inconsistency.
            if self.remote_known_whats is None:
                self.remote_known_whats = page_parser.result['legal_whats']
            if page_parser.result['title']:
                activity.title = page_parser.result['title']
            if page_parser.result['description']:
                activity.description = page_parser.result['description']
            # MMT sends different values of the current activity type, hopefully what_3 is always the
            # correct one.
            if page_parser.result['what_3']:
                activity.what = page_parser.result['what_3']
            if page_parser.result['public'] is not None:
                activity.public = page_parser.result['public']

    def load_full(self, activity):
        """get the entire activity"""
        with activity.loading():
            with MMTSession(self) as session:
                response = session.get('{}/assets/php/gpx.php?tid={}'.format(
                    self._base_url(), activity.id_in_backend))
                activity.parse(response.text)
                # but this does not give us activity type and other things,
                # get them from the web page.
                self._load_page_in_session(activity, session)

    def _remove_activity_in_backend(self, activity):
        """remove on the server"""
        act_id = activity.id_in_backend
        response = self.__post('delete_activity', activity_id=act_id)
        type_xml = response.find('type')
        if type_xml is not None and type_xml.text == 'invalid_activity_id':
            # does not exist anymore, silently ignore this.
            return
        if type_xml is None or type_xml.text != 'activity_deleted':
            raise Exception('{}: Could not delete activity {}: {}'.format(self, activity, response.text))

    def _save_full(self, activity):
        """save full gpx track on the MMT server.
        We must upload the title separately.
        Because we cannot upload the time, we set the activity time to the time
        of the first trackpoint."""

        if not activity.gpx.get_track_points_no():
            raise Exception('MMT does not accept an activity without trackpoints:{}'.format(activity))
        mmt_status = 'public' if activity.public else 'private'
        if activity.id_in_backend:
            # we cannot change an MMT activity in-place, we need to re-upload and then
            # remove the previous instance.
            self._remove_activity_in_backend(activity)
        response = self.__post(
            'upload_activity', gpx_file=activity.to_xml(),
            status=mmt_status, description=activity.description, activity=activity.what)
        activity.id_in_backend = response.find('id').text
        self._write_title(activity)

    def update(self, activity, points):
        """append points in the backend. activity already has them.
        points are GPXTrackPoint"

        Todo:
            Doc is wrong, must rethink this.
        """
        activity.add_points(points)
        self.__post(
            'update_activity', activity_id=activity.id_in_backend,
            points=points)

MMT._define_support() # pylint: disable=protected-access



#!/usr/bin/env python3

import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

long_description = """
Gpxity is a Python - library making it easy to move activities between different backends.
In this context, a backend is a place where activities can be stored.

Implemented backends are:

  * Directory          for .gpx files on an accessible file system
  * ServerDirectory    suited for a server implementation
  * MMT                for activities on http://mapmytracks.com

Sometimes you might just change a harmless thing like the description but
the backend does not allow changing this separately, so we have to re-upload
the whole activity. If it is is big and the remote server slow, this might
take 10 minutes or more. Right now this library has no asynchronous interface,
so it can really take some time until your program continues.

"""

setup(
    name='Gpxity',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    description='A uniform interface to GPX services like mapmytracks or gpsies',
    long_description=long_description,
    url='https://github.com/wrohdewald/Gpxity',
    author='Wolfgang Rohdewald',
    author_email='wolfgang@rohdewald.de',
    license='GPLv2',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Communications',
        'Topic :: Internet :: WWW/HTTP',
        ],
    packages=find_packages(exclude=['doc', 'test', 'bin', '.gitignore']),
    install_requires=['requests', 'gpxpy'],
    extras_require={
        'test': ['coverage'],
        'dev': ['sphinx', 'sphinx-autodoc-annotation']
        })



